from pathlib import Path
import sys

import pygame

from auraboros.gamescene import Scene, SceneManager
from auraboros import global_
from auraboros.global_ import init  # noqa
from auraboros.schedule import IntervalCounter
from auraboros.gametext import TextSurfaceFactory
from auraboros.utilities import AssetFilePath

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

pygame.init()

clock = pygame.time.Clock()

textfactory = TextSurfaceFactory()
textfactory.register_font(
    "misaki_gothic",
    pygame.font.Font(AssetFilePath.font("misaki_gothic.ttf"), 16))


class TitleMenuScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        textfactory.set_current_font("misaki_gothic")
        textfactory.register_text("hello_world", "Hello, World!")

    def update(self, dt):
        pass

    def draw(self, screen):
        textfactory.render("hello_world", screen, (16, 0))


def run(fps_num=60):
    global fps
    fps = fps_num
    running = True
    scene_manager = SceneManager()
    scene_manager.push(TitleMenuScene(scene_manager))
    while running:
        dt = clock.tick(fps)/1000  # dt means delta time

        global_.screen.fill((0, 0, 0))
        for event in pygame.event.get():
            running = scene_manager.event(event)

        scene_manager.update(dt)
        scene_manager.draw(global_.screen)
        pygame.transform.scale(global_.screen, global_.w_size_unscaled,
                               pygame.display.get_surface())
        pygame.display.update()
        IntervalCounter.tick(dt)
    pygame.quit()



from typing import Callable
import abc

import pygame

from .gametext import TextSurfaceFactory
from . import global_


class MenuHasNoItemError(Exception):
    pass


class GameMenuSystem:
    def __init__(self):
        self.menu_selected_index = 0
        self.menu_option_keys = []
        self.menu_option_texts = []
        self.option_actions_on_select = {}
        self.option_actions_on_highlight = {}
        self.loop_cursor = True
        self.action_on_cursor_up = lambda: None
        self.action_on_cursor_down = lambda: None

    def add_menu_item(
            self, option_key,
            action_on_select: Callable,
            action_on_highlight: Callable = lambda: None, text: str = None):
        if text is None:
            text = option_key
        self.menu_option_keys.append(option_key)
        self.menu_option_texts.append(text)
        self.option_actions_on_select[option_key] = action_on_select
        self.option_actions_on_highlight[option_key] = action_on_highlight

    def set_action_on_cursor_up(self, action: Callable):
        self.action_on_cursor_up = action

    def set_action_on_cursor_down(self, action: Callable):
        self.action_on_cursor_down = action

    def menu_cursor_up(self):
        if 0 < self.menu_selected_index:
            self.menu_selected_index -= 1
        elif self.loop_cursor:
            self.menu_selected_index = self.count_menu_items() - 1
        self.action_on_cursor_up()

    def menu_cursor_down(self):
        if self.menu_selected_index < len(self.menu_option_keys)-1:
            self.menu_selected_index += 1
        elif self.loop_cursor:
            self.menu_selected_index = 0
        self.action_on_cursor_down()

    def do_selected_action(self):
        if len(self.menu_option_keys) == 0:
            raise MenuHasNoItemError(
                "At least one menu item is required to take action.")
        return self.option_actions_on_select[
            self.menu_option_keys[self.menu_selected_index]]()

    def action_on_highlight(self):
        if len(self.menu_option_keys) == 0:
            raise MenuHasNoItemError(
                "At least one menu item is required to take action.")
        return self.option_actions_on_highlight[
            self.menu_option_keys[self.menu_selected_index]]()

    def select_action_by_index(self, index):
        if 0 <= index < len(self.menu_option_keys):
            self.menu_selected_index = index
        else:
            raise ValueError("Given index is out of range in the menu.")

    def count_menu_items(self) -> int:
        return len(self.menu_option_keys)

    def max_option_text_length(self) -> int:
        return max([len(i) for i in self.menu_option_texts])

    def update(self):
        self.action_on_highlight()


class UIElement(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def min_size(self) -> list[int]:
        self.resize_min_size_to_suit()
        return self.__min_size

    @abc.abstractmethod
    def resize_min_size_to_suit(self):
        self.__min_size = [0, 0]


class GameMenuUI(UIElement):
    """
    option_highlight_style = "cursor" or "filled_box"
    "cursor" is default
    anchor(unused) = "top_left" or "center_fixed" or "center"
    "top_left" is default
    """

    def __init__(self, menu_system: GameMenuSystem,
                 textfactory: TextSurfaceFactory,
                 option_highlight_style="cursor"):
        self.system = menu_system
        self.textfactory = textfactory
        self.resize_min_size_to_suit()
        self._pos = [0, 0]
        self.frame_color = (255, 255, 255)
        self.option_highlight_color = (222, 222, 222)
        self.option_highlight_bg_color = (122, 122, 122)
        self.cursor_size = textfactory.char_size()
        self.reposition_cursor()
        self.option_highlight_style = option_highlight_style
        self.padding = 0
        self.locate_cursor_inside_window = True
        # self.anchor = "top-left"

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, value):
        self._pos = value
        self.reposition_cursor()

    @property
    def min_size(self):
        self.resize_min_size_to_suit()
        return self.__min_size

    def resize_min_size_to_suit(self):
        self.__min_size = [
            self.system.max_option_text_length(
            )*self.textfactory.char_size()[0],
            self.system.count_menu_items()*self.textfactory.char_size()[1]]

    @property
    def ultimate_size(self):
        return self.calculate_ultimate_size()

    def calculate_ultimate_size(self) -> list[int, int]:
        if (self.option_highlight_style == "cursor" and
                self.locate_cursor_inside_window):
            size = [self.min_size[0]+self.padding*3+self.cursor_size[0],
                    self.min_size[1]+self.padding*2]
        else:
            size = [self.min_size[0]+self.padding*2,
                    self.min_size[1]+self.padding*2]
        return size

    def reposition_cursor(self):
        self.cursor_pos = [
            self.pos[0]-self.cursor_size[0],
            self.pos[1]]

    def set_x_to_center(self):
        self.pos[0] = global_.w_size[0]//2-self.ultimate_size[0]//2
        self.reposition_cursor()

    def set_y_to_center(self):
        self.pos[1] = global_.w_size[1]//2-self.ultimate_size[1]//2
        self.reposition_cursor()

    def set_pos_to_center(self):
        self.set_x_to_center()
        self.set_y_to_center()

    def draw(self, screen):
        pygame.draw.rect(
            screen, self.frame_color,
            self.pos + self.ultimate_size, 1)
        for i, (key, text) in enumerate(
            zip(self.system.menu_option_keys,
                self.system.menu_option_texts)):
            if (self.option_highlight_style == "cursor" and
                    self.locate_cursor_inside_window):
                text_pos = (
                    self.pos[0]+self.padding+self.cursor_size[0]+self.padding,
                    self.pos[1]+self.textfactory.char_size()[1]*i+self.padding)
            else:
                text_pos = (
                    self.pos[0]+self.padding,
                    self.pos[1]+self.textfactory.char_size()[1]*i+self.padding)
            self.textfactory.register_text(
                key, text, text_pos)
        if self.option_highlight_style == "cursor":
            if (self.option_highlight_style == "cursor" and
                    self.locate_cursor_inside_window):
                cursor_polygon_points = ((
                    self.cursor_pos[0]+self.cursor_size[0]+self.padding,
                    self.cursor_pos[1]+self.cursor_size[1]
                    * self.system.menu_selected_index+self.padding),
                    (self.cursor_pos[0]+self.cursor_size[0]*2+self.padding,
                     (self.cursor_pos[1]+self.cursor_size[1]//2)
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding),
                    (self.cursor_pos[0]+self.cursor_size[0]+self.padding,
                     (self.cursor_pos[1]+self.cursor_size[1])
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding))
            else:
                cursor_polygon_points = ((
                    self.cursor_pos[0],
                    self.cursor_pos[1]+self.cursor_size[1]
                    * self.system.menu_selected_index+self.padding),
                    (self.cursor_pos[0]+self.cursor_size[0],
                     (self.cursor_pos[1]+self.cursor_size[1]//2)
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding),
                    (self.cursor_pos[0],
                     (self.cursor_pos[1]+self.cursor_size[1])
                     + self.cursor_size[1]*self.system.menu_selected_index
                     + self.padding))
            pygame.draw.polygon(
                screen, self.option_highlight_color,
                cursor_polygon_points)
        elif self.option_highlight_style == "filled_box":
            pygame.draw.rect(
                screen, self.option_highlight_bg_color,
                ((self.pos[0]+self.padding, self.pos[1]+self.cursor_size[1]
                  * self.system.menu_selected_index+self.padding),
                 (self.min_size[0], self.cursor_size[1])))
            pass
        for key in self.system.menu_option_keys:
            self.textfactory.render(key, screen)


class MsgWindow(UIElement):
    """
    type_of_sizing = "min"(default) or "fixed"
    anchor(unused) = "top_left" or "center_fixed" or "center"
    "top_left" is default
    """

    def __init__(self, font: pygame.font.Font, type_of_sizing="min"):
        self.text = ""
        # self.textfactory = textfactory
        self.font = font
        self.resize_min_size_to_suit()
        # self.resize_window_to_suit_text()
        self._pos = [0, 0]
        self.frame_color = (255, 255, 255)
        self.type_of_sizing = type_of_sizing
        self._size = [0, 0]
        self.resize_on_type_of_sizing()
        self.padding = 0

    @property
    def pos(self):
        return self._pos

    @pos.setter
    def pos(self, value):
        self._pos = value

    @property
    def min_size(self):
        self.resize_min_size_to_suit()
        return self.__min_size

    def resize_min_size_to_suit(self):
        self.__min_size = list(self.font.size(self.text))

    @property
    def size(self):
        self.resize_on_type_of_sizing()
        return self._size

    @size.setter
    def size(self, value):
        self._size = value
        self.resize_on_type_of_sizing()

    def resize_on_type_of_sizing(self):
        if self.type_of_sizing == "fixed":
            if self.min_size[0] > self._size[0]:
                self._size[0] = self.min_size[0]
            if self.min_size[1] > self._size[1]:
                self._size[1] = self.min_size[1]
        elif self.type_of_sizing == "min":
            self._size = self.min_size

    @property
    def ultimate_size(self):
        return self.calculate_ultimate_size()

    def calculate_ultimate_size(self) -> list[int, int]:
        return list(map(sum, zip(self.size, [self.padding*2, self.padding*2])))

    def rewrite_text(self, text):
        self.text = text

    def set_x_to_center(self):
        self.pos[0] = global_.w_size[0]//2-self.ultimate_size[0]//2

    def set_y_to_center(self):
        self.pos[1] = global_.w_size[1]//2-self.ultimate_size[1]//2

    def set_pos_to_center(self):
        self.set_x_to_center()
        self.set_y_to_center()

    def draw(self, screen: pygame.surface.Surface):
        frame_rect = self.pos + self.ultimate_size
        pygame.draw.rect(
            screen, self.frame_color,
            frame_rect, 1)
        screen.blit(self.font.render(
            self.text, True, (255, 255, 255)),
            tuple(map(sum, zip(self.pos, [self.padding, self.padding]))))


# class UIElement:
#     def __init__(self, surface: pygame.surface.Surface = None, ):
#         self._container = None
#         self._x = 0
#         self._y = 0
#         if surface is None:
#             self.surface = pygame.surface.Surface((0, 0))
#         else:
#             self.surface = surface
#         self.rect = self.surface.get_rect()
#         self._width = self.rect.width
#         self._height = self.rect.height

#     @property
#     def x(self):
#         return self._x

#     @x.setter
#     def x(self, value):
#         self._x = value
#         self.rect.x = self._x

#     @property
#     def y(self):
#         return self._y

#     @y.setter
#     def y(self, value):
#         self._y = value
#         self.rect.y = self._y

#     @property
#     def width(self):
#         return self._width

#     @width.setter
#     def width(self, value):
#         self._width = value
#         self.rect.width = self._width
#         self.surface = pygame.surface.Surface((self.width, self.height))

#     @property
#     def height(self):
#         return self._height

#     @height.setter
#     def height(self, value):
#         self._height = value
#         self.rect.height = self._height
#         self.surface = pygame.surface.Surface((self.width, self.height))

#     @property
#     def container(self) -> Union[None, "UILayoutBase"]:
#         return self._container

#     @container.setter
#     def container(self, value: "UILayoutBase"):
#         self._container = value

#     def update(self, dt):
#         pass

#     def draw(self, screen: pygame.surface.Surface):
#         screen.blit(self.surface, self.rect)


# class UILayoutBase(UIElement):
#     def __init__(self):
#         super().__init__()
#         self.layout = list[UIElement]()
#         self.margin_top = 0
#         self.margin_bottom = 0
#         self.margin_right = 0
#         self.margin_left = 0
#         self.padding_top = 0
#         self.padding_bottom = 0
#         self.padding_right = 0
#         self.padding_left = 0
#         self.spacing = 0


# class UIBoxLayout(UILayoutBase):
#     def __init__(self):
#         super().__init__()
#         self.orientation = "vertical"  # vertical | horizontal

#     def add_ui_element(self, ui_element: UIElement):
#         self.layout.append(ui_element)

#     def stretch_to_fit_entire(self):
#         if self.orientation == "vertical":
#             height = 0
#             widths = list[int]()
#             for element in self.layout:
#                 height += element.rect.height + self.spacing
#                 widths.append(element.width)
#             self.height = height
#             self.width = max(widths)

#     def draw(self, screen: pygame.surface.Surface):
#         self.stretch_to_fit_entire()
#         if self.orientation == "vertical":
#             i = 0
#             next_y = 0
#             for element in self.layout:
#                 rect = element.rect
#                 if i < 1:
#                     next_y += rect.height + self.spacing
#                 if 1 <= i:
#                     rect.y = next_y
#                     next_y += rect.height
#                 self.surface.blit(element.surface, rect)
#                 i += 1
#         screen.blit(self.surface, self.rect)


# class UIGameText(UIElement):
#     def __init__(self, font: pygame.font.Font, text: str):
#         super().__init__()
#         self.font = font
#         self._text = text
#         self.do_reset_surface_and_rect_when_update_text = True
#         self._update_surface_and_rect_by_new_text()

#     @property
#     def text(self):
#         return self._text

#     @text.setter
#     def text(self, value: str):
#         self._text = value
#         if self.do_reset_surface_and_rect_when_update_text:
#             self._update_surface_and_rect_by_new_text()

#     def _update_surface_and_rect_by_new_text(self):
#         self.surface = self.font.render(self.text, True, (255, 255, 255))
#         self.rect = self.surface.get_rect()

#     def draw(self, screen: pygame.surface.Surface, *args, **kwargs):
#         screen.blit(self.surface, self.rect)

# uilayout = UILayout()
# uilayout.set_ui_element(UIElement(), 6, 5)
# print(uilayout.layout)



from __future__ import annotations
# from abc import ABCMeta, abstractmethod
from inspect import isclass
import math
import random
from typing import TYPE_CHECKING, Type
if TYPE_CHECKING:
    from .gamelevel import Level

from collections import UserDict
from math import sqrt

import pygame

# from .gamescene import Scene
from .utilities import Arrow, ArrowToTurnToward
from . import global_


class Entity(pygame.sprite.Sprite):
    """
    Attributes:
            gameworld (Level):
                The game level that the entity belongs to.
            entity_container (EntityList):
                The container that holds the entity.
            arrow_of_move (ArrowToTurnToward):
                The arrow of movement for the entity.
            movement_speed (int):
                The movement speed of the entity.
            image (pygame.surface.Surface):
                The image of the entity.
            rect (pygame.Rect):
                The rectangle that encloses the entity's image.
            hitbox (pygame.Rect):
                The rectangle that encloses the entity's hitbox.
            is_visible_hitbox (bool):
                Whether the hitbox of the entity is visible or not.
            is_hitbox_on_center (bool):
                Whether the hitbox of the entity is on the center of the
                entity or not.
            invincible_to_entity (bool):
                Whether the entity is invincible to other entities or not.
            x (float):
                The x-coordinate of the entity.
            y (float):
                The y-coordinate of the entity.
            angle (float):
                The angle of the entity.
            is_moving (bool):
                Whether the entity is moving or not.
            move_dest_x (int):
                The x-coordinate of the destination of the entity's movement.
            move_dest_y (int):
                The y-coordinate of the destination of the entity's movement.
            angle_to_target (float):
                The angle to the target of the entity's movement.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gameworld: Level = None
        self.entity_container: EntityList = None
        self.arrow_of_move = ArrowToTurnToward()
        self.movement_speed = 1
        self.image = pygame.surface.Surface((0, 0))
        self.rect = self.image.get_rect()
        self._hitbox = self.image.get_rect()
        self.is_visible_hitbox = False
        self.is_hitbox_on_center = True
        self.invincible_to_entity = False
        self._x = 0
        self._y = 0
        self.angle = 0
        self.is_moving = False  # this is True when move_by func called
        self.move_dest_x = None
        self.move_dest_y = None
        self.angle_to_target = 0

    @ property
    def hitbox(self):
        return self._hitbox

    @hitbox.setter
    def hitbox(self, value):
        self._hitbox = value

    def set_hitbox_x_to_rect_center_x(self):
        self.hitbox.x = self.x + (self.rect.width - self.hitbox.width) // 2

    def set_hitbox_y_to_rect_center_y(self):
        self.hitbox.y = self.y + (self.rect.height - self.hitbox.height) // 2

    @ property
    def x(self):
        return self._x

    @ x.setter
    def x(self, value):
        self._x = round(value, 2)
        self.rect.x = self._x
        if self.is_hitbox_on_center:
            self.set_hitbox_x_to_rect_center_x()
        else:
            self.hitbox.x = self._x

    @ property
    def y(self):
        return self._y

    @ y.setter
    def y(self, value):
        self._y = round(value, 2)
        self.rect.y = self._y
        self.hitbox.y = self._y
        if self.is_hitbox_on_center:
            self.set_hitbox_y_to_rect_center_y()
        else:
            self.hitbox.y = self._y

    def set_move_direction(self, direction: Arrow):
        """
        Set the arrow of movement to the specified direction.

        Args:
            direction (Arrow):
                The direction to set the arrow of movement to.
        """
        self.arrow_of_move.set(direction)

    def cancel_move_direction(self, direction: Arrow):
        """
        Unset the arrow of movement from the specified direction.

        Args:
            direction (Arrow):
                The direction to unset the arrow of movement from.
        """
        self.arrow_of_move.unset(direction)

    def move_by_arrow(self):
        """
        Move the entity according to the arrow of movement.

        If the arrow of movement is set to up, down, right, or left,
        the entity moves in that direction at the speed of movement.
        If the arrow of movement is set to up and right, up and left,
        down and right, or down and left, the entity moves diagonally
        at the speed of movement divided by the square root of 2.

        The entity's position is updated accordingly, and the is_moving
        attribute is set to True if the entity is moving, or False
        otherwise.
        """
        if ((self.arrow_of_move.is_up and
            self.arrow_of_move.is_right) or
            (self.arrow_of_move.is_up and
            self.arrow_of_move.is_left) or
            (self.arrow_of_move.is_down and
            self.arrow_of_move.is_right) or
            (self.arrow_of_move.is_down and
                self.arrow_of_move.is_left)):
            # Correct the speed of diagonal movement
            movement_speed = self.movement_speed / sqrt(2)
        else:
            movement_speed = self.movement_speed
        # movement_speed = movement_speed * dt * global_.TARGET_FPS
        movement_speed = movement_speed
        if self.arrow_of_move.is_up:
            self.y -= movement_speed
        if self.arrow_of_move.is_down:
            self.y += movement_speed
        if self.arrow_of_move.is_right:
            self.x += movement_speed
        if self.arrow_of_move.is_left:
            self.x -= movement_speed
        if self.arrow_of_move.is_set_any():
            self.is_moving = True
        else:
            self.is_moving = False

    def move_by_angle(self, dt, radians):
        self.is_moving = True
        self.y += math.sin(radians)*self.movement_speed
        self.x += math.cos(radians)*self.movement_speed

    def set_x_to_center_of_screen(self):
        """Center the posistion on the screen"""
        self.x = global_.w_size[0] / 2 - self.rect.width / 2

    def set_y_to_center_of_screen(self):
        """Center the posistion on the screen"""
        self.y = global_.w_size[1] / 2 - self.rect.height / 2

    def remove_from_container(self):
        self.entity_container.kill_living_entity(self)

    def death(self):
        self.remove_from_container()

    @staticmethod
    def collide(entity_a: Entity, entity_b: Entity,
                death_a=True, death_b=True) -> bool:
        """Each entity executes death() when a collision occurs."""
        is_entity_a_alive = entity_a in entity_a.gameworld.entities
        is_entity_b_alive = entity_a in entity_b.gameworld.entities
        if (entity_a.hitbox.colliderect(entity_b.hitbox)
            and entity_b.hitbox.colliderect(entity_a.hitbox)
                and is_entity_a_alive and is_entity_b_alive):
            # if (collided(entity_a, entity_b)
            #         and is_entity_a_alive and is_entity_b_alive):
            if not entity_a.invincible_to_entity:
                if death_a:
                    entity_a.death()
            if not entity_b.invincible_to_entity:
                if death_b:
                    entity_b.death()
            return True
        else:
            return False

    def draw(self, screen: pygame.surface.Surface):
        screen.blit(self.image, self.rect)
        if self.is_visible_hitbox:
            self.draw_hitbox(screen)

    def draw_hitbox(self, screen: pygame.surface.Surface):
        screen.fill((255, 0, 0), self.hitbox)

    def visible_hitbox(self):
        self.is_visible_hitbox = True

    def invisible_hitbox(self):
        self.is_visible_hitbox = False

    def set_arrow_random_vertical(self, dt):
        """Set arrow of movement to right or left."""
        if not self.move_dest_x:
            self.random_dest_x()
        if (self.move_dest_x - self.movement_speed
            <= self.x <=
                self.move_dest_x + self.movement_speed):
            self.arrow_of_move.unset(Arrow.right)
            self.arrow_of_move.unset(Arrow.left)
            self.random_dest_x()
        elif self.x < self.move_dest_x:
            self.arrow_of_move.set(Arrow.right)
            self.arrow_of_move.unset(Arrow.left)
        elif self.move_dest_x < self.x:
            self.arrow_of_move.set(Arrow.left)
            self.arrow_of_move.unset(Arrow.right)

    def set_arrow_random_horizontal(self, dt):
        """Set arrow of movement to up or down."""
        if not self.move_dest_y:
            self.random_dest_y()
        if (self.move_dest_y - self.movement_speed
            <= self.y <=
                self.move_dest_y + self.movement_speed):
            self.arrow_of_move.unset(Arrow.up)
            self.arrow_of_move.unset(Arrow.down)
            self.random_dest_x()
        elif self.y < self.move_dest_y:
            self.arrow_of_move.set(Arrow.down)
            self.arrow_of_move.unset(Arrow.up)
        elif self.move_dest_y < self.y:
            self.arrow_of_move.set(Arrow.down)
            self.arrow_of_move.unset(Arrow.up)

    def set_arrow_random(self, dt):
        """Set arrow of movement to up or down, right, or left."""
        v_or_h = random.randint(0, 1)
        if v_or_h == 0:
            self.set_arrow_random_vertical(dt)
        elif v_or_h == 1:
            self.set_arrow_random_horizontal(dt)

    def random_dest_x(self):
        self.move_dest_x = random.randint(0, global_.w_size[0])

    def random_dest_y(self):
        self.move_dest_y = random.randint(0, global_.w_size[1])

    def set_arrow_to_entity_as_dest(self, dt, entity_type: Entity):
        if self.set_destination_to_entity(entity_type):
            self.set_arrow_to_dest(dt)

    def set_arrow_to_dest_x(self, dt, stop_when_arrived=True):
        if self.move_dest_x:
            if (self.move_dest_x - self.movement_speed
                + self.hitbox.width
                <= self.x <=
                    self.move_dest_x + self.movement_speed
                    - self.hitbox.width) and stop_when_arrived:
                self.arrow_of_move.unset(Arrow.right)
                self.arrow_of_move.unset(Arrow.left)
            elif self.x < self.move_dest_x:
                self.arrow_of_move.set(Arrow.right)
                self.arrow_of_move.unset(Arrow.left)
            elif self.move_dest_x < self.x:
                self.arrow_of_move.set(Arrow.left)
                self.arrow_of_move.unset(Arrow.right)

    def set_arrow_to_dest_y(self, dt):
        if self.move_dest_y:
            if (self.move_dest_y - self.movement_speed
                + self.hitbox.height
                <= self.y <=
                    self.move_dest_y + self.movement_speed
                    - self.hitbox.height):
                self.arrow_of_move.unset(Arrow.up)
                self.arrow_of_move.unset(Arrow.down)
            elif self.y < self.move_dest_y:
                self.arrow_of_move.set(Arrow.down)
                self.arrow_of_move.unset(Arrow.up)
            elif self.move_dest_y < self.y:
                self.arrow_of_move.set(Arrow.up)
                self.arrow_of_move.unset(Arrow.down)

    def set_arrow_to_dest(self, dt):
        self.set_arrow_to_dest_x(dt)
        self.set_arrow_to_dest_y(dt)

    def set_destination_to_entity(self, entity_type: Entity) -> bool:
        entity_list = [
            entity for entity in self.gameworld.entities
            if isinstance(entity, entity_type)]
        if entity_list:
            self.move_dest_x = entity_list[0].hitbox.x
            self.move_dest_y = entity_list[0].hitbox.y
            return True
        else:
            return False


class EntityList(list[Entity]):
    def kill_living_entity(self, entity: Entity):
        """Do list.remove(entity) if list has it."""
        if entity in self:
            self.remove(entity)

    def append(self, item: Entity):
        if not isinstance(item, Entity):
            raise TypeError("item is not Entity")
        super().append(item)


class ShooterEntity(Entity):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shot_max_num = 1
        self.shot_interval = 1
        self.is_shot_allowed = True


class DeadlyObstacle(Entity):
    pass


class Enemy(DeadlyObstacle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.action = "idle"
        self.behavior_pattern = None
        self.behavior_pattern_dict = {}
        self.behavior_pattern_dict[
            "random_vertical"] = self.set_arrow_random_vertical
        self.behavior_pattern_dict[
            "random_horizontal"] = self.set_arrow_random_horizontal
        self.behavior_pattern_dict[
            "random"] = self.set_arrow_random
        self.gamescore = 0

    def update(self, dt):
        self.do_pattern(dt)

    def do_pattern(self, dt):
        if self.behavior_pattern is not None:
            self.behavior_pattern_dict[self.behavior_pattern](dt)
            self.move_by_arrow(dt)


class EnemyFactory(UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, item: Enemy):
        if isclass(item):
            self.data[key] = item
        else:
            raise TypeError("The value must not be instance.")

    def __getitem__(self, key) -> Type[Enemy]:
        return super().__getitem__(key)



import pytest
from unittest.mock import MagicMock
import moderngl
from src.auraboros.shader import Shader2D


@pytest.fixture
def mock_ctx(mocker):
    mock_ctx = MagicMock(spec=moderngl.Context)
    mocker.patch('moderngl.create_context', return_value=mock_ctx)
    return mock_ctx


def test_shader_singleton(mock_ctx):
    shader1 = Shader2D()
    shader2 = Shader2D()
    assert shader1 is shader2
    shader1.test_attr = "test"
    assert shader2.test_attr is shader1.test_attr



from dataclasses import dataclass, field, asdict
from pathlib import Path
import json
import sys

import pygame

from auraboros import global_


def open_json_file(filepath):
    with open(filepath, "r") as f:
        return json.load(f)


@dataclass
class Arrow:
    """Arrow symbol"""
    LEFT = 0
    UP = 1
    RIGHT = 2
    DOWN = 3


@dataclass
class ArrowToTurnToward:
    """Use to set direction"""
    is_up: bool = field(default=False)
    is_down: bool = field(default=False)
    is_right: bool = field(default=False)
    is_left: bool = field(default=False)

    def set(self, direction: Arrow):
        if direction is Arrow.UP:
            self.is_up = True
        elif direction is Arrow.DOWN:
            self.is_down = True
        elif direction is Arrow.RIGHT:
            self.is_right = True
        elif direction is Arrow.LEFT:
            self.is_left = True

    def unset(self, direction: Arrow):
        if direction is Arrow.UP:
            self.is_up = False
        elif direction is Arrow.DOWN:
            self.is_down = False
        elif direction is Arrow.RIGHT:
            self.is_right = False
        elif direction is Arrow.LEFT:
            self.is_left = False

    # def switch(self, direction: Arrow):
    #     if direction is Arrow.up:
    #         self.is_down = False
    #     elif direction is Arrow.down:
    #         self.is_down = not self.is_down
    #     elif direction is Arrow.right:
    #         self.is_right = self.is_right
    #     elif direction is Arrow.left:
    #         self.is_left = self.is_left

    def is_set_any(self):
        return True in set(asdict(self).values())


class AssetFilePath:
    root_dirname = "assets"
    root_dir_parent = Path(sys.argv[0]).parent
    __root = root_dir_parent / root_dirname
    root = Path(__root)
    img_dirname = "imgs"
    font_dirname = "fonts"
    sound_dirname = "sounds"

    @ classmethod
    def pyinstaller_path(cls, filepath):
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            path = Path(sys._MEIPASS) / cls.root_dirname / filepath
            # path will be such as: "sys._MEIPASS/assets/imgs/example.png"
        except AttributeError:
            path = cls.root / filepath
        return path

    @ classmethod
    def img(cls, filename):
        return cls.pyinstaller_path(Path(cls.img_dirname) / filename)

    @ classmethod
    def font(cls, filename):
        return cls.pyinstaller_path(Path(cls.font_dirname) / filename)

    @ classmethod
    def sound(cls, filename):
        return cls.pyinstaller_path(Path(cls.sound_dirname) / filename)

    @ classmethod
    def set_asset_root(cls, root_dir_path: str):
        cls.__root = root_dir_path
        cls.root = Path(cls.__root)
        cls.root_dir_parent = Path(root_dir_path).parent
        cls.root_dirname = Path(root_dir_path).name


def draw_grid_background(
        screen: pygame.surface.Surface, grid_size: int, color: int):
    [pygame.draw.rect(
        screen, color,
        (x*grid_size, y*grid_size) + (grid_size, grid_size), 1)
        for x in range(global_.w_size[0]//grid_size)
        for y in range(global_.w_size[1]//grid_size)]



from dataclasses import dataclass
from typing import Any, Tuple

import pygame

from . import global_

pygame.font.init()


@dataclass
class GameText:
    text: str
    pos: list
    rgb: list
    surface: pygame.surface.Surface = None

    def draw(self, screen: pygame.surface.Surface):
        screen.blit(self.surface, self.pos)


class TextSurfaceFactory:
    def __init__(self):
        self.current_font_key = None
        self._text_dict: dict[Any, GameText] = {}
        self._font_dict = FontDict()

    @property
    def text_dict(self) -> dict[Any, GameText]:
        return self._text_dict

    @property
    def font_dict(self):
        return self._font_dict

    def register_text(self, key, text: str = "", pos=[0, 0],
                      color_rgb=[255, 255, 255]):
        self.text_dict[key] = GameText(text=text, pos=pos, rgb=color_rgb)

    def rewrite_text(self, key, text: str):
        self.text_dict[key].text = text

    def text_by_key(self, key) -> str:
        return self.text_dict[key].text

    def is_text_registered(self, key):
        if self.text_dict.get(key):
            return True
        else:
            return False

    def register_font(self, key, font: pygame.font.Font):
        if len(self.font_dict) == 0:
            self.current_font_key = key
        self.font_dict[key] = font

    def set_current_font(self, key):
        self.current_font_key = key

    def font_by_key(self, key) -> pygame.font.Font:
        return self.font_dict[key]

    def font(self) -> pygame.font.Font:
        """Return Font object that is currently being used"""
        return self.font_dict[self.current_font_key]

    def char_size(self) -> Tuple[int, int]:
        return self.font_dict[self.current_font_key].size(" ")

    def set_text_pos(self, key, pos):
        self.text_dict[key].pos = pos

    def set_text_color(self, key, color_rgb):
        self.text_dict[key].rgb = color_rgb

    def set_text_pos_to_right(self, key):
        self.text_dict[key].pos[0] = \
            global_.w_size[0] - \
            self.font().size(self.text_dict[key].text)[0]

    def set_text_pos_to_bottom(self, key):
        self.text_dict[key].pos[1] = \
            global_.w_size[1] - \
            self.font().size(self.text_dict[key].text)[1]

    def center_text_pos_x(self, key):
        self.text_dict[key].pos[0] = \
            global_.w_size[0]//2 - \
            self.font().size(self.text_dict[key].text)[0]//2

    def center_text_pos_y(self, key):
        self.text_dict[key].pos[1] = \
            global_.w_size[1]//2 - \
            self.font().size(self.text_dict[key].text)[1]//2

    def render(self, text_key, surface_to_draw: pygame.surface.Surface,
               pos=None, wait_rendering_for_text_to_register=True):
        if self.is_text_registered(text_key):
            text_surf = self.font().render(
                self.text_by_key(text_key), True,
                self.text_dict[text_key].rgb)
            if pos is None:
                pos_ = self.text_dict[text_key].pos
            else:
                pos_ = pos
            self.text_dict[text_key].surface = text_surf
            surface_to_draw.blit(text_surf, pos_)

    def generate_gametext(self, text_key):
        return self.text_dict[text_key]


class FontDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value: pygame.font.Font):
        if isinstance(value, pygame.font.Font):
            super().__setitem__(key, value)
        else:
            raise TypeError("The value must be Font object of pygame.")

    def __getitem__(self, key) -> pygame.font.Font:
        return super().__getitem__(key)



from array import array
from pathlib import Path
from typing import Any
import os


import moderngl
import pygame

# from . import global_

with open(Path(__file__).parent / "default.vert", "r") as f:
    VERTEX_DEFAULT = f.read()
with open(Path(__file__).parent / "default.frag", "r") as f:
    FRAGMENT_DEFAULT = f.read()


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class Shader2D(metaclass=Singleton):
    """2Dシェーダーを表すシングルトンクラス。

    Attributes:
        ctx (moderngl.Context): ModernGLのコンテキストオブジェクト。
        textures (dict): テクスチャオブジェクトを格納する辞書。
        programs (dict): シェーダープログラムオブジェクトを格納する辞書。
        vaos (dict): 頂点配列オブジェクトを格納する辞書。
        buffer (moderngl.Buffer): 頂点バッファオブジェクト。
    """

    def __init__(self):
        self.ctx = moderngl.create_context()
        self.textures: dict[Any, moderngl.Texture] = {}
        self.programs: dict[Any, moderngl.Program] = {}
        self.vaos: dict[Any, moderngl.VertexArray] = {}
        self.buffer: dict[Any, moderngl.Buffer] = self.ctx.buffer(
            data=array("f", [
                # x, y, u ,v
                -1.0, 1.0, 0.0, 0.0,  # top left
                1.0, 1.0, 1.0, 0.0,  # top right
                -1.0, -1.0, 0.0, 1.0,  # bottom left
                1.0, -1.0, 1.0, 1.0,  # bottom right
            ]))

        self.compile_and_register_program(
            vertex=VERTEX_DEFAULT,
            fragment=FRAGMENT_DEFAULT,
            program_name="display_surface")

    def compile_and_register_program(self, vertex, fragment, program_name):
        """

        Args:
            vertex (str): 頂点シェーダーのソースコード。
            fragment (str): フラグメントシェーダーのソースコード。
            program_name (str): シェーダープログラムの名前。
        """
        vert_raw = vertex
        frag_raw = fragment
        program = self.ctx.program(
            vertex_shader=vert_raw, fragment_shader=frag_raw)
        self.programs[program_name] = program
        self.vaos[program_name] = self.ctx.vertex_array(
            program, [(self.buffer, "2f 2f", "in_vert", "in_texcoord")])

    def _surface_to_texture(self, surface: pygame.surface.Surface):
        texture = self.ctx.texture(surface.get_size(), 4)
        texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        if os.name == 'nt':
            # Windows
            texture.swizzle = 'BGRA'
        return texture

    def register_surface_as_texture(
            self, surface: pygame.surface.Surface, texture_name):
        """PygameのSurfaceオブジェクトをテクスチャとして登録する。

        Args:
            surface (pygame.surface.Surface): PygameのSurfaceオブジェクト。
            texture_name (str): テクスチャの名前。

        Notes:
            テクスチャが辞書に登録されていない場合は、
            PygameのSurfaceオブジェクトからテクスチャオブジェクトを作成し、
            辞書に登録する。
            テクスチャオブジェクトにSurfaceオブジェクトのデータを書き込む。
        """
        if texture_name not in self.textures:
            texture = self._surface_to_texture(surface)
            self.textures[texture_name] = texture
        buffer = surface.get_view("1")
        self.textures[texture_name].write(buffer)

    def use_texture(self, texture_name, id):
        """texture_nameでtextures辞書に登録されたテクスチャを指定し、使用を切り替えます。
        renderメソッドを実行した際、これで指定したテクスチャに描画されます。

        Args:
            texture_name (str): テクスチャの名前。
            id (int): テクスチャのID。
        """
        self.textures[texture_name].use(id)

    def set_uniform(self, program_name, uniform, value):
        """シェーダープログラムのuniformに値を設定する。

        Args:
            program_name (str): シェーダープログラムの名前。
            uniform (str): uniform変数の名前。
            value (any): uniform変数に設定する値。
        """
        self.programs[program_name][uniform].value = value

    def render(self):
        for program_name in self.programs:
            self.vaos[program_name].render(mode=moderngl.TRIANGLE_STRIP)




from collections import deque
from pathlib import Path
import sys
from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.utilities import AssetFilePath
from auraboros.gametext import TextSurfaceFactory
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros import global_

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

textfactory = TextSurfaceFactory()
textfactory.register_font(
    "misaki_gothic",
    pygame.font.Font(AssetFilePath.font("misaki_gothic.ttf"), 16))

QWERTY_STR = "qwertyuiopasdfghjklzxcvbnm"
AZERTY_STR = "azertyuiopqsdfghjklmwxcvbn"


class KeyboardDebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textinput = ""
        self.textinput_to_show = self.textinput
        textfactory.set_current_font("misaki_gothic")
        self.keyboard["qwerty"] = Keyboard()
        self.keyboard["azerty"] = Keyboard()
        for key_name in QWERTY_STR:
            self.keyboard["qwerty"].register_keyaction(
                pygame.key.key_code(key_name), 0, 0,
                lambda kn=key_name: self.press_key(kn),
                lambda kn=key_name: self.release_key(kn))
        self.keyboard["qwerty"].register_keyaction(
            pygame.K_1, 0, 0,
            lambda: self.switch_keyboard_layout("azerty", "1"))
        for key_name in AZERTY_STR:
            self.keyboard["azerty"].register_keyaction(
                pygame.key.key_code(key_name), 0, 0,
                lambda kn=key_name: self.press_key(kn),
                lambda kn=key_name: self.release_key(kn))
        self.keyboard["azerty"].register_keyaction(
            pygame.K_2, 0, 0,
            lambda: self.switch_keyboard_layout("qwerty", "2"))
        self.key_i_o_map: dict[str: bool] = dict.fromkeys(
            ascii_lowercase, False)
        self.keyboard.set_current_setup("qwerty")

    def press_key(self, key):
        self.textinput += key
        self.key_i_o_map[key] = True
        # print(key)
        textfactory.register_text("recent_pressed", f"{key}")
        textfactory.register_text("textinput", f"{self.textinput_to_show}")

    def release_key(self, key):
        self.key_i_o_map[key] = False
        # print(key)
        textfactory.register_text("recent_pressed", f"{key}")

    def switch_keyboard_layout(self, layout_name, key):
        print(key)
        self.keyboard.set_current_setup(layout_name)

    def update(self, dt):
        for key_name in ascii_lowercase:
            self.keyboard.current_setup.do_action_by_keyinput(
                pygame.key.key_code(key_name))
        self.keyboard.current_setup.do_action_by_keyinput(
            pygame.K_1, True)
        self.keyboard.current_setup.do_action_by_keyinput(
            pygame.K_2, True)

        textfactory.register_text(
            "current_layout",
            f"layout:{self.keyboard.current_setup_key}",
            color_rgb=(178, 150, 250))

    def draw(self, screen):
        if self.keyboard.current_setup_key == "qwerty":
            keyboard_layout = QWERTY_STR
        elif self.keyboard.current_setup_key == "azerty":
            keyboard_layout = AZERTY_STR
        char_size = textfactory.font().size("a")
        for i, key_name in enumerate(keyboard_layout):
            if i < 10:  # key_name <= "p"(qwerty)
                surface_pos = (i*char_size[0], 0)
            elif 10 <= i < 19:  # key_name <= "l"(qwerty)
                surface_pos = (char_size[0]//3+(i-10)
                               * char_size[0], char_size[1])
            elif 19 <= i:  # key_name <= "m"(qwerty)
                surface_pos = (char_size[0]//2+(i-19)
                               * char_size[0], char_size[1]*2)
            if self.key_i_o_map[key_name]:
                text_surface = textfactory.font().render(
                    key_name, True, (89, 255, 89))
                screen.blit(text_surface, surface_pos)
            else:
                text_surface = textfactory.font().render(
                    key_name, True, (255, 255, 255))
                screen.blit(text_surface, surface_pos)
        textfactory.render("current_layout", screen, (0, char_size[1]*4))
        textinput_size = textfactory.font().size(self.textinput)
        textinput_homepos = (0, char_size[1]*6)
        if textinput_size[0] > global_.w_size[0]:
            num_of_chars = global_.w_size[0] // char_size[0]
            textinput_lines = [
                self.textinput[i:i+num_of_chars]
                for i in range(0, len(self.textinput), num_of_chars)]
            if (textinput_size[1]*len(textinput_lines) >
                    global_.w_size[1]-textinput_homepos[1]):
                textinput_deque = deque(self.textinput)
                self.textinput = textinput_deque.pop()
        else:
            textinput_lines = [self.textinput]
        for line_num, line in enumerate(textinput_lines):
            textinput_to_show = line
            textinput_surface = textfactory.font().render(
                textinput_to_show, True, (255, 255, 255))
            screen.blit(
                textinput_surface,
                (textinput_homepos[0],
                 textinput_homepos[1]+char_size[1]*line_num))


scene_manager = SceneManager()
scene_manager.push(KeyboardDebugScene(scene_manager))

if __name__ == "__main__":
    engine.init()
    engine.run(scene_manager=scene_manager)



from pathlib import Path
import subprocess

import click


@click.command()
@click.option('--example', is_flag=True, required=False,
              help="navigate to choose example scripts.")
def cli(example):
    if example:
        example_dir = Path(__file__).parent.parent / "debugs"
        example_scripts = [f for f in example_dir.glob(
            "*.py") if not f.name == "init_for_dev.py"]
        click.echo(f"Here are {len(example_scripts)} examples:")
        for i, file_name in enumerate(example_scripts):
            click.echo(f"{i} {file_name.name} ({file_name})")
        example_num = click.prompt(
            "Choose an example to try:", type=int, default=0)
        subprocess.run(["python", example_scripts[example_num]])
    else:
        click.echo(click.get_current_context().get_help())


if __name__ == '__main__':
    cli()



from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Type
if TYPE_CHECKING:
    pass
    # from .eightrail import Enemy

from random import randint

import copy

import pygame

from .entity import Enemy, EnemyFactory, Entity, EntityList, DeadlyObstacle
from .gamescene import Scene
from .utilities import Arrow, open_json_file
from . import global_


class EntityListOfGameWorld(EntityList):
    def __init__(self, gameworld: Level, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gameworld = gameworld

    def kill_living_entity(self, entity: Entity):
        if isinstance(entity, Enemy):
            self.gameworld.num_of_remaining_enemies -= 1
        super().kill_living_entity(entity)

    def append(self, item: Entity):
        if isinstance(item, Enemy):
            self.gameworld.count_of_enemies_summoned += 1
        item.gameworld = self.gameworld
        item.entity_container = self
        super().append(item)


class Level:

    def __init__(self, levelfile_dir, scene=None):
        self.scene: Scene = scene

        self.levelfile_dir = levelfile_dir
        self.level_raw_data = open_json_file(self.levelfile_dir / "level.json")
        self.level_raw_pattern_data = open_json_file(
            levelfile_dir / "patterns.json")
        self.prepare_level_data()

        self._entities = EntityListOfGameWorld(self)
        self.enemy_factory = EnemyFactory()

        self.bg_surf = pygame.surface.Surface(
            (global_.w_size[0], global_.w_size[1] * 2))
        self.scroll_speed = 0.5
        self.density_of_stars_on_bg = randint(100, 500)

        self.pause = False
        self.elapsed_time_in_level = 0
        self.count_of_enemies_killed = 0
        self.count_of_enemies_summoned = 0
        self.reset_num_of_remaining_enemies()

        self.gamescore: int
        self.scoreboard = [0, ]

        self.initialize_level()

    @property
    def entities(self):
        return self._entities

    @entities.setter
    def entities(self, value):
        self._entities = value

    def reset_num_of_remaining_enemies(self):
        self.num_of_remaining_enemies = self.num_of_enemy_on_level()

    def reset_count_of_enemies_summoned(self):
        self.count_of_enemies_summoned = 0

    def entity(self, entity_type: Type[Entity]):
        """Return the entity of specified type which added first to
           entities."""
        entity_list = [
            entity for entity in self.entities
            if isinstance(entity, entity_type)]
        if entity_list:
            return entity_list[0]

    def entities_by_type(self, entity_type: Type[Entity]):
        entity_list = [
            entity for entity in self.entities
            if isinstance(entity, entity_type)]
        if entity_list:
            return entity_list

    def show_hitbox(self):
        self._visible_hitbox()

    def hide_hitbox(self):
        self._invisible_hitbox()

    def _visible_hitbox(self):
        [entity.visible_hitbox() for entity in self.entities]

    def _invisible_hitbox(self):
        [entity.invisible_hitbox() for entity in self.entities]

    def highscore(self):
        self.scoreboard.sort(reverse=True)
        return self.scoreboard[0]

    def enemies(self) -> EntityList[Enemy]:
        enemy_list = [
            enemy for enemy in self.entities if isinstance(enemy, Enemy)]
        return enemy_list

    def deadly_obstacles(self) -> EntityList[DeadlyObstacle]:
        deadly_obstacle_list = [
            deadly_obstacle for deadly_obstacle in self.entities
            if isinstance(deadly_obstacle, DeadlyObstacle)]
        return deadly_obstacle_list

    def num_of_enemy_now(self):
        return len(self.enemies())

    def prepare_level_data(self):
        level = []
        pattern_dict = self.level_raw_pattern_data
        for data in self.level_raw_data:
            decompressed = pattern_dict[data[0]]
            list_ = []
            for dict_ in decompressed:
                new_dict = copy.copy(dict_)
                new_dict["timing"] = dict_["timing"]+data[1]
                list_.append(new_dict)
            level.extend(list_)
        self.level_data = level

    def run_level(self, dt):
        if self.pause:
            return

        for data in self.level_data:
            if round(self.elapsed_time_in_level) == data["timing"]:
                enemy = self.enemy_factory[data["enemy"]]()
                pos: list[int, int] = [None, None]
                for i in range(2):
                    if isinstance(data["pos"][i], str):
                        if data["pos"][i] == "random":
                            pos[i] = randint(0, global_.w_size[i])
                        if data["pos"][i] == "right":
                            pos[i] = global_.w_size[i] - enemy.rect.width
                    else:
                        pos[i] = data["pos"][i]
                enemy.x, enemy.y = pos
                enemy.behavior_pattern = data["pattern"]
                self.entities.append(enemy)

        self.elapsed_time_in_level += 1 * dt * global_.TARGET_FPS

    def process_collision(
            self,
            player_entities: Iterable[Entity],
            weapon_entities: Iterable[Entity]) -> bool:
        for deadly_obstacle in self.deadly_obstacles():
            for weapon in weapon_entities:
                if isinstance(deadly_obstacle, Enemy):
                    if Entity.collide(weapon, deadly_obstacle):
                        self.gamescore += deadly_obstacle.gamescore
                        self.count_of_enemies_killed += 1
                else:
                    Entity.collide(weapon, deadly_obstacle, False)
            for player in player_entities:
                Entity.collide(player, deadly_obstacle)

    def register_gamescore(self):
        self.scoreboard.append(self.gamescore)

    def read_tagged_level_data(self):
        data_dict_by_tag = {}
        for data in self.level_raw_data:
            if isinstance(data, list):
                without_str = list(filter(
                    lambda item: not isinstance(item, list),
                    self.level_raw_data))
                data_dict_by_tag[data[0]] = [
                    item for item in without_str if item["tag"] == data[0]]
        return data_dict_by_tag

    def num_of_enemy_on_level(self) -> int:
        return len(self.level_data)

    def kill_count_of_enemy(self) -> int:
        return self.count_of_enemies_killed

    # def num_of_remaining_enemy(self) -> int:
    #     return

    def reset_elapsed_time_counter(self):
        self.elapsed_time_in_level = 0

    def clear_enemies(self):
        [enemy.remove_from_container() for enemy in self.enemies()]

    def clear_entities(self):
        self.entities.clear()

    def reset_scroll(self):
        self.bg_scroll_y = 0
        self.bg_scroll_x = 0

    def summon_enemies_with_timing_resetted(self):
        self.clear_enemies()
        self.reset_elapsed_time_counter()

    def reset_score(self):
        self.gamescore = 0

    def initialize_level(self):
        self.clear_entities()
        self.reset_elapsed_time_counter()
        self.reset_scroll()
        self.reset_score()
        self.reset_num_of_remaining_enemies()
        self.reset_count_of_enemies_summoned()

    def clear_enemies_off_screen(self):
        [entity.remove_from_container() for entity in self.enemies()
         if global_.w_size[1] < entity.y or global_.w_size[0] < entity.x
            or entity.x < 0 or entity.y < 0]

    def stop_entity_from_moving_off_screen(self, entity: Entity):
        if global_.w_size[1] < entity.y + entity.rect.height:
            entity.arrow_of_move.unset(Arrow.down)
        if global_.w_size[0] < entity.x + entity.rect.width:
            entity.arrow_of_move.unset(Arrow.right)
        if entity.x < 0:
            entity.arrow_of_move.unset(Arrow.left)
        if entity.y < 0:
            entity.arrow_of_move.unset(Arrow.up)

    def set_background(self):
        [self.bg_surf.fill(
            (randint(0, 255), randint(0, 255), randint(0, 255)),
            ((randint(0, global_.w_size[0]),
              randint(0, global_.w_size[1] * 2)), (1, 1)))
         for i in range(self.density_of_stars_on_bg)]

    def set_background_for_scroll(self):
        new_background = pygame.surface.Surface(
            (global_.w_size[0], global_.w_size[1] * 2))
        new_background.blit(
            self.bg_surf,
            (0, global_.w_size[1], global_.w_size[0], global_.w_size[1]))
        randomize_density = randint(-self.density_of_stars_on_bg // 2,
                                    self.density_of_stars_on_bg // 2)
        [new_background.fill(
            (randint(0, 255), randint(0, 255), randint(0, 255)),
            ((randint(0, global_.w_size[0]), randint(0, global_.w_size[1])),
             (1, 1)))
         for i in range(self.density_of_stars_on_bg + randomize_density)]
        self.bg_surf = new_background

    def scroll(self, dt):
        for enemy in self.enemies():
            enemy.y += self.scroll_speed * 1.25 * dt * global_.TARGET_FPS
        self.bg_scroll_y += self.scroll_speed * dt * global_.TARGET_FPS
        if self.bg_scroll_y > global_.w_size[1]:
            self.bg_scroll_y = 0
            self.set_background_for_scroll()




# from collections import deque
from pathlib import Path
import sys
# from string import ascii_lowercase

import pygame

import init_for_dev  # noqa
from auraboros import engine
from auraboros.utilities import AssetFilePath, draw_grid_background
from auraboros.gametext import TextSurfaceFactory
from auraboros.gamescene import Scene, SceneManager
from auraboros.gameinput import Keyboard
from auraboros.ui import GameMenuSystem, GameMenuUI, MsgWindow
from auraboros.animation import AnimationImage, SpriteSheet
from auraboros import global_

engine.init(caption="Test Animation System")

AssetFilePath.set_asset_root(Path(sys.argv[0]).parent / "assets")

textfactory = TextSurfaceFactory()
textfactory.register_font(
    "misaki_gothic",
    pygame.font.Font(AssetFilePath.font("misaki_gothic.ttf"), 16))


class TestAnimImg(AnimationImage):
    def __init__(self):
        super().__init__()
        self.sprite_sheet = SpriteSheet(AssetFilePath.img("testsprite.png"))
        self.anim_frames: list[pygame.surface.Surface] = [
            self.sprite_sheet.image_by_area(0, 0, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*4, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*3, 32, 32),
            self.sprite_sheet.image_by_area(0, 32*2, 32, 32),
            self.sprite_sheet.image_by_area(0, 32, 32, 32)]
        self.anim_interval = 1000
        self.loop_count = 1


class DebugScene(Scene):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        textfactory.set_current_font("misaki_gothic")
        self.test_anim_img = TestAnimImg()
        self.keyboard["menu"] = Keyboard()
        self.keyboard.set_current_setup("menu")
        self.menusystem = GameMenuSystem()
        self.keyboard["menu"].register_keyaction(
            pygame.K_UP, 0, 122, self.menusystem.menu_cursor_up)
        self.keyboard["menu"].register_keyaction(
            pygame.K_DOWN, 0, 122, self.menusystem.menu_cursor_down)
        self.keyboard["menu"].register_keyaction(
            pygame.K_z, 0, 0, self.menusystem.do_selected_action)
        self.menusystem.add_menu_item(
            "play", self.play_animation, text="Play")
        self.menusystem.add_menu_item(
            "stop", self.stop_animation, text="STOP")
        self.menusystem.add_menu_item(
            "reset", self.reset_animation, text="RESET")
        self.menuui = GameMenuUI(self.menusystem, textfactory, "filled_box")
        self.menuui.padding = 4
        self.msgbox = MsgWindow(textfactory.font())
        self.msgbox.padding = 4
        self.msgbox.text = "Press 'Z'"
        self.msgbox2 = MsgWindow(textfactory.font())
        self.msgbox2.padding = 4
        self.msgbox3 = MsgWindow(textfactory.font())
        self.msgbox3.padding = 4
        self.msgbox4 = MsgWindow(textfactory.font())
        self.msgbox4.padding = 4
        self.msgbox5 = MsgWindow(textfactory.font())
        self.msgbox5.padding = 4
        self.msgbox6 = MsgWindow(textfactory.font())
        self.msgbox6.padding = 4
        self.msgbox7 = MsgWindow(textfactory.font())
        self.msgbox7.padding = 4
        self.elapsed_time = 0
        self.start_time = 0
        self.stopwatch_is_running = False

    def play_animation(self):
        self.test_anim_img.let_play()
        if not self.stopwatch_is_running:
            self.start_time = pygame.time.get_ticks()
        self.stopwatch_is_running = True

    def stop_animation(self):
        self.test_anim_img.let_stop()
        self.stopwatch_is_running = False

    def reset_animation(self):
        self.test_anim_img.reset_animation()
        self.elapsed_time += pygame.time.get_ticks() - self.start_time
        self.start_time = pygame.time.get_ticks()
        self.elapsed_time = 0

    def update(self, dt):
        if self.stopwatch_is_running:
            self.elapsed_time = pygame.time.get_ticks() - self.start_time
        if self.test_anim_img.is_all_loop_finished():
            self.stopwatch_is_running = False
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_UP)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_DOWN)
        self.keyboard.current_setup.do_action_by_keyinput(pygame.K_z)
        self.menuui.set_pos_to_center()
        self.menusystem.update()
        self.msgbox2.text = \
            f"loop_count:{self.test_anim_img.loop_count}"
        self.msgbox3.text = \
            f"anim_interval:{self.test_anim_img.anim_interval} milliseconds"
        self.msgbox4.text = \
            f"anim_frame_id:{self.test_anim_img.anim_frame_id}"
        self.msgbox5.text = \
            f"is_playing:{self.test_anim_img.is_playing}"
        self.msgbox6.text = \
            f"loop_counter:{self.test_anim_img.loop_counter}"
        self.msgbox7.text = \
            f"elapsed time:{self.elapsed_time/1000}"
        self.msgbox2.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1]
        self.msgbox3.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1]
        self.msgbox4.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1]
        self.msgbox5.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1]
        self.msgbox6.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1]
        self.msgbox7.pos[1] = \
            self.msgbox.calculate_ultimate_size()[1] +\
            self.msgbox2.calculate_ultimate_size()[1] +\
            self.msgbox3.calculate_ultimate_size()[1] +\
            self.msgbox4.calculate_ultimate_size()[1] +\
            self.msgbox5.calculate_ultimate_size()[1] +\
            self.msgbox6.calculate_ultimate_size()[1]

    def draw(self, screen):
        draw_grid_background(screen, 16, (78, 78, 78))
        screen.blit(
            self.test_anim_img.image,
            (global_.w_size[0]//2-self.test_anim_img.image.get_width()//2,
             self.menuui.pos[1]-self.test_anim_img.image.get_height()))
        self.menuui.draw(screen)
        self.msgbox.draw(screen)
        self.msgbox2.draw(screen)
        self.msgbox3.draw(screen)
        self.msgbox4.draw(screen)
        self.msgbox5.draw(screen)
        self.msgbox6.draw(screen)
        self.msgbox7.draw(screen)


scene_manager = SceneManager()
scene_manager.push(DebugScene(scene_manager))

if __name__ == "__main__":
    engine.run(scene_manager=scene_manager, fps=60)



"""
Use to define global variables common use in the modules.
"""
import pygame

pygame.init()


screen: pygame.surface.Surface
TARGET_FPS: int
w_size_unscaled: tuple[int, int]
w_size: tuple[int, int]


def init(window_size=(960, 640), caption="", icon_filepath=None,
         pixel_scale=2, set_mode_flags=0):
    """This function initialize pygame and game engine.
    Where to configure settings of game system is here."""
    from . import global_
    global_.TARGET_FPS = 60
    pixel_scale = pixel_scale
    global_.w_size_unscaled = window_size
    global_.w_size = tuple([length // pixel_scale for length in window_size])
    pygame.display.set_mode(global_.w_size_unscaled, set_mode_flags)
    global_.screen = pygame.Surface(global_.w_size)
    pygame.display.set_caption(caption)
    if icon_filepath:
        icon_surf = pygame.image.load(icon_filepath)
        pygame.display.set_icon(icon_surf)



# from dataclasses import dataclass
from collections import UserDict
from inspect import isclass
from typing import Any, TypedDict, Union

import pygame


# @dataclass
# class SoundManager:
#     sound_dict = dict[Any, pygame.mixer.Sound] = {}


class SoundDict(UserDict):

    def __init__(self, channel: pygame.mixer.Channel = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.channel = channel
        self.sound = None

    def __setitem__(self, key, value: pygame.mixer.Sound):
        if not isclass(value):
            self.data[key] = value
        else:
            raise ValueError("The value must be instance.")


class ChannelManager(UserDict):
    def register(
            self, key: Any, channel: pygame.mixer.Channel,
            volume: Union[float, tuple[float, float]]):
        channel.set_volume(volume)
        self.data[key]: ChannelDict = ChannelDict(
            channel=channel, volume=volume)


class _ChannelDictData(TypedDict):
    channel: pygame.mixer.Channel
    volume: Union[float, tuple[float, float]]


class ChannelDict(UserDict):
    def __init__(self, dict=None, /, **kwargs):
        self.data: _ChannelDictData = _ChannelDictData()
        if dict is not None:
            self.update(dict)
        if kwargs:
            self.update(kwargs)

    def __setitem__(self, key, value):
        if key == "volume":
            if not (value < 0 or 1.0 < value):
                self.data["channel"].set_volume(value)
            else:
                return
        super().__setitem__(key, value)
